import argparse

from collections import OrderedDict

from zope.dottedname.resolve import resolve

import aio.app
import aio.config
import aio.signals
import aio.app.logging


def start_listening(app, signals=None):
    app.signals = signals or aio.signals.Signals()


def load_modules(app, config):
    # load up builtins and modules
    app.modules = []

    try:
        _modules = [
            m.strip() for
            m in config['aio']['builtin'].strip('').split('\n')
            if m.strip()]
        for m in _modules:
            app.modules.append(resolve(m))
    except KeyError:
        pass

    try:
        _modules = [
            m.strip() for
            m in config['aio']['modules'].strip('').split('\n')
            if m.strip()]
        for m in _modules:
            app.modules.append(resolve(m))
    except KeyError:
        pass

    app.modules = tuple(app.modules)


def setup_config(app, config_file=None, config_string=None,
                 search_for_config=None):

    import aio.config

    # read default aio.app.config
    config = aio.config.parse_config(
        modules=[app])

    # read user config
    config = aio.config.parse_config(
        config=config_file,
        config_string=config_string,
        parser=config,
        search_for_config=search_for_config)

    load_modules(app, config)

    # read module config
    config = aio.config.parse_config(
        modules=app.modules, parser=config)

    # read user config again
    app.config = aio.config.parse_config(
        config=config_file,
        config_string=config_string,
        parser=config,
        search_for_config=search_for_config)


def get_commands(app):
    commands = OrderedDict(
        app.config['aio/builtin_commands'])
    if "aio/commands" in app.config:
        commands.update(
            OrderedDict(app.config['aio/commands']))
    return commands


def run_command(command, argv):
    return resolve(command)(argv)


def runner(argv, app=None, configfile=None,
           signals=None, config_string=None, search_for_config=False):
    app = app or resolve("aio.app")
    parser = argparse.ArgumentParser(
        prog="aio",
        description='aio app runner')
    parser.add_argument(
        "-c", nargs="?",
        help="configuration file")
    parsed = parser.parse_known_args(argv)

    if parsed[0].c:
        configfile = parsed[0].c
    setup_config(
        app,
        config_file=configfile,
        config_string=config_string,
        search_for_config=search_for_config)
    commands = get_commands(app)
    parser.add_argument(
        "command", choices=commands,
        help="command to run")
    try:
        parsed_args, remainder = parser.parse_known_args(argv)
    except (SystemExit, IndexError):
        parser.print_help()
        return
    except Exception as e:
        # import traceback
        # traceback.print_exc()
        print(repr(e))
        return

    if parsed_args.command in commands:
        start_listening(app, signals)
        run_command(commands[parsed_args.command], remainder)
    else:
        parser.print_help()



import asyncio
import logging

from zope.dottedname.resolve import resolve

import aio.app


def listener(*la, **kwa):

    def wrapper(func):

        def wrapped(*la, **kwa):
            if asyncio.iscoroutinefunction(func):
                coro = func
            else:
                coro = asyncio.coroutine(func)
            try:
                result = yield from coro(*la)
            except Exception as e:
                # import traceback
                # traceback.print_exc()
                logging.getLogger('aio').error(
                    "Error calling listener: %s" % e)
                raise e
            return result
        wrapped.__annotations__['decorator'] = listener
        wrapped.__annotations__['function'] = func
        return wrapped

    if len(la) == 1 and callable(la[0]):
        return wrapper(la[0])
    return wrapper


def listener_checker(func):
    if not func.__annotations__.get("decorator") == listener:
        dotted_func = "%s.%s" % (func.__module__, func.__name__)
        raise RuntimeError(
            "Signal listener (%s) must be wrapped with " % dotted_func
            + "@aio.app.signal.listener")


def start_listeners(config):
    log = logging.getLogger("aio")
    log.debug('adding event listeners')

    try:
        listener_check = config["aio/signals"]["listener_check"]
        listener_check = resolve(listener_check)
    except (IndexError, KeyError):
        listener_check = None

    for s in config.sections():
        if s.startswith("listen/"):
            section = config[s]
            for signal, handlers in section.items():
                for handler in [h.strip() for h in handlers.split('\n')]:
                    handler = resolve(handler)
                    if listener_check:
                        listener_check(handler)
                    aio.app.signals.listen(signal, handler)



import asyncio
import logging
import inspect

from zope.dottedname.resolve import resolve

from aio.core.exceptions import MissingConfiguration
import aio.app


def factory(*la, **kwa):

    def wrapper(func):

        def wrapped(*la, **kwa):
            if asyncio.iscoroutinefunction(func):
                coro = func
            else:
                coro = asyncio.coroutine(func)
            try:
                result = yield from coro(*la)
            except Exception as e:
                # import traceback
                # traceback.print_exc()
                logging.getLogger('aio').error(
                    "Error starting server: %s" % e)
                raise e
            return result
        wrapped.__annotations__['decorator'] = factory
        wrapped.__annotations__['function'] = func
        return wrapped

    if len(la) == 1 and callable(la[0]):
        return wrapper(la[0])
    return wrapper


def factory_checker(func):
    if not func.__annotations__.get("decorator") == factory:
        dotted_func = "%s.%s" % (func.__module__, func.__name__)
        raise RuntimeError(
            "Server factory (%s) must be wrapped with " % dotted_func
            + "@aio.app.server.factory")


def protocol(*la, **kwa):

    def wrapper(func):

        def wrapped(*la, **kwa):
            if asyncio.iscoroutinefunction(func):
                coro = func
            else:
                coro = asyncio.coroutine(func)
            try:
                result = yield from coro(*la)
            except Exception as e:
                # import traceback
                # traceback.print_exc()
                logging.getLogger('aio').error(
                    "Error starting server: %s" % e)
                raise e
            return result
        wrapped.__annotations__['decorator'] = protocol
        wrapped.__annotations__['function'] = func
        return wrapped

    if len(la) == 1 and callable(la[0]):
        return wrapper(la[0])
    return wrapper


def protocol_checker(func):
    if inspect.isclass(func) and issubclass(func, asyncio.Protocol):
        return

    if not func.__annotations__.get("decorator") == protocol:
        dotted_func = "%s.%s" % (func.__module__, func.__name__)
        raise RuntimeError(
            "Server factory (%s) must be wrapped with " % dotted_func
            + "@aio.app.server.factory")


@asyncio.coroutine
def server_factory(name, protocol, address, port):
    loop = asyncio.get_event_loop()
    return (
        yield from loop.create_server(
            protocol, address, port))


@asyncio.coroutine
def start_server(name, address="127.0.0.1", port=8080,
                 factory=None, protocol=None):
    log = logging.getLogger("aio")

    if not port:
        raise MissingConfiguration(
            "Section [server/%s] must specify port to listen on" % name)

    if not factory and not protocol:
        message = (
            "Section [server/%s] must specify one of factory or "
            + "protocol to start server") % name
        raise MissingConfiguration(message)

    if not factory:
        factory = server_factory

    if "function" in factory.__annotations__:
        original_function = factory.__annotations__['function']
    else:
        original_function = factory
    module = "%s.%s" % (
        original_function.__module__,
        original_function.__name__)

    try:
        res = yield from factory(name, protocol, address, port)
    except Exception as e:
        log.error("Server(%s) failed to start: %s" % (name, module))
        import traceback
        traceback.print_exc()
        raise e

    aio.app.servers[name] = res
    log.info(
        'Server(%s) %s started on %s:%s' % (
            name,
            module,
            address or "*", port))
    return res


@asyncio.coroutine
def start_servers(config):
    log = logging.getLogger("aio")
    log.debug('adding servers')
    try:
        factory_check = config["aio/server"]["factory_check"]
        factory_check = resolve(factory_check)
    except (IndexError, KeyError):
        factory_check = None

    try:
        protocol_check = config["aio/server"]["protocol_check"]
        protocol_check = resolve(protocol_check)
    except (IndexError, KeyError):
        protocol_check = None

    for s in config.sections():
        if s.startswith("server/"):
            name = s.split("/")[1].strip()
            section = config[s]
            factory = section.get('factory', None)
            if factory:
                factory = resolve(factory)
                if factory_check:
                    factory_check(factory)
            protocol = section.get('protocol', None)
            if protocol:
                protocol = resolve(protocol)
                if protocol_check:
                    protocol_check(protocol)
            address = section.get('address')
            port = section.get('port')

            log.debug("Starting server: %s" % name)

            task = asyncio.async(
                aio.app.server.start_server(
                    name, address, port, factory, protocol))

            def _server_started(res):
                if res.exception():
                    loop = asyncio.get_event_loop()
                    loop.stop()
                    log.error(res.exception())
                    raise res.exception()

            task.add_done_callback(_server_started)



import asyncio
import logging

from zope.dottedname.resolve import resolve

import aio.app


class ScheduledEvent(object):

    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name

    @property
    def config(self):
        return aio.app.config["schedule/%s" % self.name]


def schedule(name, func, cb, t, exc=None):
    log = logging.getLogger("aio")
    log.info(
        'Scheduler started (%s): %s.%s' % (
            name, func.__module__, func.__name__))
    while True:
        if not asyncio.iscoroutinefunction(func):
            func = asyncio.coroutine(func)
        future = asyncio.async(func(ScheduledEvent(name)))

        def _cb(res):
            if res.exception():
                if exc:
                    asyncio.async(exc(res.exception()))
            else:
                if cb:
                    asyncio.async(cb(res.result()))

        future.add_done_callback(_cb)
        yield from asyncio.sleep(t)


@asyncio.coroutine
def start_schedulers(config):
    log = logging.getLogger("aio")
    log.debug('adding schedulers')
    for s in config.sections():
        if s.startswith("schedule/"):
            msg = s.split("/")[1].strip()
            section = config[s]
            every = section.get("every")
            func = resolve(section.get('func'))

            cb = section.get('cb', None)
            if cb:
                cb = resolve(cb)

            err = section.get('err', None)
            if err:
                err = resolve(err)

            log.debug("Starting scheduler: %s" % msg)
            asyncio.async(
                aio.app.schedule.schedule(
                    msg, func, cb, int(every), err))



import os
import asyncio
import logging
import argparse
import configparser
import lockfile

from daemon import runner, daemon

import aio.app
import aio.config


def cmd_config(argv):

    parser = argparse.ArgumentParser(
        prog="aio config",
        description='aio configuration')

    parser.add_argument(
        "-f",
        nargs=1,
        help=(
            "Configuration file to get/set values from. "
            + "Values are not interpolated if you set this"))

    parser.add_argument(
        "-g", "--get",
        nargs=1,
        help="Get a config value")

    parser.add_argument(
        "-s", "--set",
        nargs=2,
        help="Set a config value")

    parsed = parser.parse_args(argv)

    if parsed.set and parsed.get:
        parser.print_help()
        raise Exception(
            "Please use either -g (--get) or -s (--set) "
            + "with aio config, not both")

    if not parsed.set and not parsed.get:
        yield from aio.config.dump_config(aio.app.config)
    else:
        if parsed.get:
            if parsed.f:
                config = configparser.RawConfigParser()
                config.read_file(open(parsed.f[0]))
            else:
                config = aio.app.config

            if ":" in parsed.get[0]:
                section = parsed.get[0].split(":")[0]
                option = parsed.get[0].split(":")[1]
            else:
                section = parsed.get[0]
                option = None
            if option:
                print(config[section][option])
            else:
                for option_name, option in config[section].items():
                    print(
                        "%s = %s" % (
                            option_name,
                            option.replace("\n", "\n\t")))
        else:
            k, v = parsed.set

            if ":" not in k:
                parser.print_help()
                raise Exception(
                    'Please specify a "section:key" to set the value for')

            conf_file = parsed.f or aio.config.find_config()
            if not conf_file:
                conf_file = os.path.abspath('aio.conf')

            config_parser = configparser.RawConfigParser()
            config_parser.read_file(open(conf_file))
            section = k.split(":")[0]
            option = k.split(":")[1]

            if section not in config_parser.sections():
                config_parser.add_section(section)

            # not sure if this is the correct way to unescape str
            v = bytes(v, 'utf-8').decode('unicode-escape')

            # now lets replace any newlines with
            # a newline + tab to ensure its treated as multi-line
            v.replace("\n", "\n\t")
            config_parser[section][option] = v
            config_parser.write(open(conf_file, "w"))
            log = logging.getLogger('aio')
            log.info("Config file (%s) updated" % conf_file)


@asyncio.coroutine
def app_start():
    from aio.app import config
    log = logging.getLogger('aio')
    log.debug('aio app starting')
    yield from aio.app.signals.emit('aio-starting', None)
    yield from aio.app.schedule.start_schedulers(config)
    yield from aio.app.server.start_servers(config)
    yield from aio.app.signals.emit('aio-started', None)
    log.debug('aio app started')
    return 3


def app_runner():
    loop = asyncio.get_event_loop()
    asyncio.set_event_loop(loop)
    asyncio.async(app_start())
    if not loop.is_running():
        loop.run_forever()


def cmd_run(argv):
    parser = argparse.ArgumentParser(
        prog="aio run",
        description='aio app runner')
    parser.add_argument(
        "-d", action="store_true",
        help="daemonize process")
    parsed, remainder = parser.parse_known_args()

    aio.app.logging.start_logging(aio.app.config)
    aio.app.signal.start_listeners(aio.app.config)

    if parsed.d:
        pidfile = runner.make_pidlockfile(
            os.path.abspath('var/run/aio.pd'), 1)

        if runner.is_pidfile_stale(pidfile):
            pidfile.break_lock()

        if pidfile.is_locked():
            print(
                "There seems to be another aio process running "
                + "already, stop that one first")
            exit()

        class open_logs:

            def __enter__(self):
                return (
                    open("var/log/aio.log", "a"),
                    open("var/log/aio.log", "a"))

            def __exit__(self, *la):
                pass

        with open_logs() as (stdout, stderr):
            daemon_context = dict(
                stdout=stdout,
                stderr=stderr,
                working_directory=os.getcwd(),
                pidfile=pidfile)
            dc = daemon.DaemonContext(**daemon_context)
            try:
                dc.open()
                app_runner()
            except lockfile.AlreadyLocked:
                print('LOCKFILE LOCKED')
    else:
        app_runner()
